from pathlib import Path

import boto3
from transformers import (
    TrainingArguments,
    TrainerCallback,
    TrainerControl,
    TrainerState
)


class S3UploadCallback(TrainerCallback):
    def __init__(self, bucket_name, model_prefix):
        self.bucket_name = bucket_name
        self.model_prefix = model_prefix
        self.s3 = boto3.client("s3")

    def on_save(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs
    ):
        if state.is_world_process_zero:
            model_dir = Path(args.output_dir)
            for file_path in model_dir.glob('*'):
                s3_key = f"{self.model_prefix}/{file_path.name}"
                self.s3.upload_file(str(file_path), self.bucket_name, s3_key)



def tokenize_and_process_dataset(dataset, teacher_tokenizer):
    def process(examples):
        tokenized_inputs = teacher_tokenizer(
            examples["sentence"], truncation=True, max_length=512
        )
        return tokenized_inputs

    tokenized_datasets = dataset.map(process, batched=True)
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")

    labels = tokenized_datasets["train"].features["labels"].names
    num_labels = len(labels)
    label2id, id2label = dict(), dict()

    for id, label in enumerate(labels):
        label2id[label] = str(id)
        id2label[str(id)] = label

    return tokenized_datasets, num_labels, label2id, id2label



def benchmark(sentences, base_model, tiny_model):
    base_correct = tiny_correct = 0
    total = len(sentences)


    for sentence, correct in sentences:
        print(sentence)
        correct_sentence = sentence.replace('[MASK]', correct).lower()

        base_prediction = base_model(sentence)[0]['sequence']
        tiny_prediction = tiny_model(sentence)[0]['sequence']

        print(tiny_prediction)
        print(base_prediction)

        base_correct += int(base_prediction == correct_sentence)
        tiny_correct += int(tiny_prediction == correct_sentence)
    
    return tiny_correct / total, base_correct / total



import torch
import torch.nn as nn
import torch.nn.functional as F

from transformers import Trainer, TrainingArguments, PreTrainedModel


class DistillationTrainingArguments(TrainingArguments):
    def __init__(self, *args, alpha=0.5, temperature=2.0, **kwargs):
        super().__init__(*args, **kwargs)

        self.alpha = alpha
        self.temperature = temperature


class DistillationTrainer(Trainer):
    def __init__(self, *args, teacher_model=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.teacher: PreTrainedModel = teacher_model  # type: ignore
        
        self._move_model_to_device(self.teacher,self.model.device)
        self.teacher.eval()

    def compute_loss(self, model, inputs, return_outputs=False):
        labels = inputs.get('labels')
        
        # compute student output
        outputs_student = model(**inputs, labels=labels)
        student_loss=outputs_student.loss
        # compute teacher output
        with torch.no_grad():
          outputs_teacher = self.teacher(**inputs)

        # Soften probabilities and compute distillation loss
        loss_function = nn.KLDivLoss(reduction="batchmean")
        loss_logits = (loss_function(
            F.log_softmax(outputs_student.logits / self.args.temperature, dim=-1),
            F.softmax(outputs_teacher.logits / self.args.temperature, dim=-1)) * (self.args.temperature ** 2))
        # Return weighted student loss
        loss = self.args.alpha * student_loss + (1. - self.args.alpha) * loss_logits
        return (loss, outputs_student) if return_outputs else loss